#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2008 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Security.Policy;
using System.Reflection;
using System.Diagnostics;
using System.Security.Permissions;
using System.Security;
using System.Linq;

namespace HeuristicLab.PluginInfrastructure.Manager {

  public class ApplicationManager : MarshalByRefObject, IApplicationManager {
    private List<IPluginDescription> plugins;
    /// <summary>
    /// Gets all plugins.
    /// </summary>
    public IEnumerable<IPluginDescription> Plugins {
      get { return plugins; }
    }

    private List<IApplicationDescription> applications;
    /// <summary>
    /// Gets all installed applications.
    /// </summary>
    public IEnumerable<IApplicationDescription> Applications {
      get { return applications; }
    }

    public ApplicationManager() : base() { }

    internal void PrepareApplicationDomain(IEnumerable<IApplicationDescription> apps, IEnumerable<IPluginDescription> plugins) {
      this.plugins = new List<IPluginDescription>(plugins);
      this.applications = new List<IApplicationDescription>(apps);
      PluginInfrastructure.ApplicationManager.RegisterApplicationManager(this);
      LoadPlugins(plugins);
    }

    private void LoadPlugins(IEnumerable<IPluginDescription> plugins) {
      // load all loadable plugins (all dependencies available) into the execution context
      foreach (var desc in PluginDescriptionIterator.IterateInDependencyOrder(plugins.Where(x => x.PluginState != PluginState.Disabled))) {
        foreach (var plugin in GetInstances<IPlugin>(desc)) {
          plugin.OnLoad();
        }
        desc.Load();
      }
    }

    internal void Run(IApplicationDescription appInfo) {
      IApplication runnablePlugin = (IApplication)Activator.CreateInstance(appInfo.DeclaringAssemblyName, appInfo.DeclaringTypeName).Unwrap();
      try {
        runnablePlugin.Run();
      }
      catch (Exception e) {
        throw new Exception(String.Format(
          "Unexpected exception caught: \"{0}\"\r\n" +
          "Type: {1}\r\n" +
          "Plugin {2}:\r\n{3}",
          e.Message,
          e.GetType().FullName,
          appInfo.Name,
          e.ToString()));
      }
    }

    /// <summary>
    /// Creates an instance of all types that are subtypes or the same type of the specified type and declared in <paramref name="plugin"/>
    /// </summary>
    /// <typeparam name="T">Most general type.</typeparam>
    /// <returns>Enumerable of the created instances.</returns>
    public IEnumerable<T> GetInstances<T>(IPluginDescription plugin) where T : class {
      return from t in GetTypes(typeof(T), plugin)
             where !t.IsAbstract && !t.IsInterface && !t.HasElementType
             select (T)Activator.CreateInstance(t);
    }
    /// <summary>
    /// Creates an instance of all types that are subtypes or the same type of the specified type
    /// </summary>
    /// <typeparam name="T">Most general type.</typeparam>
    /// <returns>Enumerable of the created instances.</returns>
    public IEnumerable<T> GetInstances<T>() where T : class {
      return from i in GetInstances(typeof(T))
             select (T)i;
    }

    /// <summary>
    /// Creates an instance of all types that are subtypes or the same type of the specified type
    /// </summary>
    /// <typeparam name="type">Most general type.</typeparam>
    /// <returns>Enumerable of the created instances.</returns>
    public IEnumerable<object> GetInstances(Type type) {
      return from t in GetTypes(type)
             where !t.IsAbstract && !t.IsInterface && !t.HasElementType
             select Activator.CreateInstance(t);
    }

    /// <summary>
    /// Finds all types that are subtypes or equal to the specified type.
    /// </summary>
    /// <param name="type">Most general type for which to find matching types.</param>
    /// <returns>Enumerable of the discovered types.</returns>
    public IEnumerable<Type> GetTypes(Type type) {
      return from asm in AppDomain.CurrentDomain.GetAssemblies()
             from t in GetTypes(type, asm)
             select t;
    }

    /// <summary>
    /// Finds all types that are subtypes or equal to the specified type if they are part of the given
    /// <paramref name="plugin"/>.
    /// </summary>
    /// <param name="type">Most general type for which to find matching types.</param>
    /// <param name="plugin">The plugin the subtypes must be part of.</param>
    /// <returns>Enumerable of the discovered types.</returns>
    public IEnumerable<Type> GetTypes(Type type, IPluginDescription pluginDescription) {
      return from asm in AppDomain.CurrentDomain.GetAssemblies()
             where pluginDescription.Assemblies.Contains(asm.Location)
             from t in GetTypes(type, asm)
             select t;
    }

    /// <summary>
    /// Gets types that are assignable (same of subtype) to the specified type only from the given assembly.
    /// </summary>
    /// <param name="type">Most general type we want to find.</param>
    /// <param name="assembly">Assembly that should be searched for types.</param>
    /// <returns>Enumerable of the discovered types.</returns>
    private IEnumerable<Type> GetTypes(Type type, Assembly assembly) {
      return from t in assembly.GetTypes()
             where type.IsAssignableFrom(t)
             select t;
    }


    // infinite lease time
    /// <summary>
    /// Initializes the life time service with infinite lease time.
    /// </summary>
    /// <returns><c>null</c>.</returns>
    public override object InitializeLifetimeService() {
      return null;
    }
  }
}

