#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2008 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Text;
using System.Reflection;
using System.Linq;

namespace HeuristicLab.PluginInfrastructure {
  /// <summary>
  /// Provides convenience methods to find specific types or to create instances of types.
  /// </summary>
  public class DiscoveryService {
    //internal static IEnumerable<IPluginDescription> plugins;

    //public IEnumerable<IPluginDescription> Plugins {
    //  get {
    //    return DiscoveryService.plugins;
    //  }
    //}

    /// <summary>
    /// Creates an instance of all types that are subtypes or the same type of the specified type and declared in <paramref name="plugin"/>
    /// </summary>
    /// <typeparam name="T">Most general type.</typeparam>
    /// <returns>Enumerable of the created instances.</returns>
    public IEnumerable<T> GetInstances<T>(IPluginDescription plugin) where T : class {
      return from t in GetTypes(typeof(T), plugin)
             where !t.IsAbstract && !t.IsInterface && !t.HasElementType
             select (T)Activator.CreateInstance(t);
    }
    /// <summary>
    /// Creates an instance of all types that are subtypes or the same type of the specified type
    /// </summary>
    /// <typeparam name="T">Most general type.</typeparam>
    /// <returns>Enumerable of the created instances.</returns>
    public IEnumerable<T> GetInstances<T>() where T : class {
      return from i in GetInstances(typeof(T))
             select (T)i;
    }

    /// <summary>
    /// Creates an instance of all types that are subtypes or the same type of the specified type
    /// </summary>
    /// <typeparam name="type">Most general type.</typeparam>
    /// <returns>Enumerable of the created instances.</returns>
    public IEnumerable<object> GetInstances(Type type) {
      return from t in GetTypes(type)
             where !t.IsAbstract && !t.IsInterface && !t.HasElementType
             select Activator.CreateInstance(t);
    }

    /// <summary>
    /// Finds all types that are subtypes or equal to the specified type.
    /// </summary>
    /// <param name="type">Most general type for which to find matching types.</param>
    /// <returns>Enumerable of the discovered types.</returns>
    public IEnumerable<Type> GetTypes(Type type) {
      return from asm in AppDomain.CurrentDomain.GetAssemblies()
             from t in GetTypes(type, asm)
             select t;
    }

    /// <summary>
    /// Finds all types that are subtypes or equal to the specified type if they are part of the given
    /// <paramref name="plugin"/>.
    /// </summary>
    /// <param name="type">Most general type for which to find matching types.</param>
    /// <param name="plugin">The plugin the subtypes must be part of.</param>
    /// <returns>Enumerable of the discovered types.</returns>
    public IEnumerable<Type> GetTypes(Type type, IPluginDescription pluginDescription) {
      return from asm in AppDomain.CurrentDomain.GetAssemblies()
             where pluginDescription.Assemblies.Contains(asm.Location)
             from t in GetTypes(type, asm)
             select t;
    }

    ///// <summary>
    ///// Gets instances of all types that implement the specified interface only in the given assembly.
    ///// </summary>
    ///// <typeparam name="T">Interface type.</typeparam>
    ///// <param name="assembly">Assembly that should be searched for types.</param>
    ///// <returns>The found instances as array.</returns>
    //internal T[] GetInstances<T>(Assembly assembly) {
    //  Type[] types = GetTypes(typeof(T), assembly);
    //  List<T> instances = new List<T>();
    //  foreach (Type t in types) {
    //    if (!t.IsAbstract && !t.IsInterface && !t.HasElementType) {
    //      instances.Add((T)Activator.CreateInstance(t));
    //    }
    //  }
    //  return instances.ToArray();
    //}

    /// <summary>
    /// Gets types that are assignable (same of subtype) to the specified type only from the given assembly.
    /// </summary>
    /// <param name="type">Most general type we want to find.</param>
    /// <param name="assembly">Assembly that should be searched for types.</param>
    /// <returns>Enumerable of the discovered types.</returns>
    private IEnumerable<Type> GetTypes(Type type, Assembly assembly) {
      return from t in assembly.GetTypes()
             where type.IsAssignableFrom(t)
             select t;
    }

    ///// <summary>
    ///// Gets the plugin of the given <paramref name="type"/>.
    ///// </summary>
    ///// <param name="type"></param>
    ///// <returns>The found plugin or <c>null</c>.</returns>
    //public PluginInfo GetDeclaringPlugin(Type type) {
    //  foreach (PluginInfo info in PluginManager.Manager.LoadedPlugins) {
    //    if (info.Assemblies.Contains(type.Assembly.Location)) return info;
    //  }
    //  return null;
    //}
  }
}
